local addon = LibStub("AceAddon-3.0"):NewAddon("FreeRefills","AceEvent-3.0","AceConsole-3.0")

local _G = _G
_G.FreeRefills = addon

local L = LibStub("AceLocale-3.0"):GetLocale("FreeRefills")
local LBU = LibStub("LibBagUtils-1.0")


local floor,ceil,min,max = floor,ceil,min,max
local select,type,pairs,ipairs,next = select,type,pairs,ipairs,next
local tonumber,tostring = tonumber,tostring
local strmatch,strfind,format = strmatch,strfind,format
local assert=assert
local tinsert,sort=table.insert,table.sort
local GetTime=GetTime
-- GLOBALS: LibStub
-- GLOBALS: BANK_CONTAINER
-- GLOBALS: GetItemInfo,GetItemCount,GetContainerNumFreeSlots,PickupContainerItem,SplitContainerItem
-- GLOBALS: GetMerchantNumItems,GetMerchantItemInfo,BuyMerchantItem



function addon:OnInitialize()
	local defaults = {
		profile = {
			overstock = false,
			merchant = true,
			bank = true,
			guildbank = true,
			refills = {},
		}
	}

	self.db = LibStub("AceDB-3.0"):New("FreeRefillsDB", defaults --[[, char default]])

	local options
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("FreeRefills",function()
		if not options then	-- avoid table creation until necessary
			options = {	
				type = 'group',
				name="FreeRefills",	-- displayname in GUIs
				childGroups="tab",
				
				handler = self,
				set="SetDB",
				get="GetDB",
				
				args = {
					cmdlinetools  = {
						type="group",
						name="cmdlinetools",
						hidden = true,
						cmdHidden = false,
						inline = true,
						args = {
							add = {
								order=1,
								name = L["Add"], type = 'input',
							    desc = L["Add an item to the refill list. (Usage: /freerefills add [Item Link] #)"],
								get = false,
							    set = "AddRefillItem",
								usage = L["[Item Link] <number of items>"],
								guiHidden = true
							},
							addstack = {
								order=2,
								name = L["Add Stack"], type = 'input',
							    desc = L["Add an item to the refill list by stack. (Usage: /freerefills addstack [Item Link] #)"],
							    get = false,
								set = "AddRefillItemStack",
								usage = L["[Item Link] <number of stacks>"],
								guiHidden = true
							},
							del = {
								order=3,
								name = L["Remove"], type = 'input',
							    desc = L["Remove an item from the refill list. (Usage: /freerefills del [Item Link])"],
								get = false,
							    set = "RemoveRefillItem",
								usage = L["[Item Link]"],
								guiHidden = true
							},
							list = {
								order=4,
								name = L["List"], type = 'execute',
							    desc = L["List the items in the refill list."],
							    func = "List",
							},
						},
					},
					gui = {
						cmdHidden = true,
						type="description",
						name = "Use the command line to manage the list of items to be refilled. Use \"/freerefills help\"",
					},
					overstock = {
						order=200,
						name = L["Overstock"], type = 'toggle',
						desc = L["Toggle Overstocking.  If an item is sold in stacks and cannot meet your exact requested amount, you can opt to buy slightly more (Overstock)."],
					},
					merchant = {
						order=201,
						name = L["Process Merchant"], type = 'toggle',
						desc = L["Toggle Merchant Looting."],
					},
					bank = {
						order=202,
						name = L["Process Bank"], type = 'toggle',
						desc = L["Toggle Bank Looting."],
					},
					guildbank = {
						order=203,
						name = L["Process Guildbank"], type = 'toggle',
						desc = L["Toggle Guildbank Looting."],
					},
					reset = {
						order=299,
						name = L["Clear"], type = 'execute',
					    desc = L["Clear the refill list completely."],
					    func = "ClearRefills",
					},
					profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db),
				}
			}
		end
		
		return options
	end)
	
	self:RegisterChatCommand("freerefills", "OnChatCommand", true)
	self:RegisterChatCommand("fr", "OnChatCommand", true)
end

addon.onSets={}
function addon:SetDB(info,v)
	local name = info[#info]
	self.db.profile[name]=v
	if self.onSets[name] then
		self[self.onSets[name]](self,v)
	end
end

function addon:GetDB(info)
	return self.db.profile[info[#info]]
end

function addon:OnChatCommand(input)
	LibStub("AceConfigCmd-3.0"):HandleCommand("freerefills", "FreeRefills", input)
end

addon.onSets["merchant"]="OnSet_merchant"
function addon:OnSet_merchant(v)
	self[v and "RegisterEvent" or "UnregisterEvent"](self,"MERCHANT_SHOW")
	self:Printf("Will %s refill from merchants.", v and "now" or "no longer")
end

addon.onSets["bank"]="OnSet_bank"
function addon:OnSet_bank(v)
	self[v and "RegisterEvent" or "UnregisterEvent"](self,"BANKFRAME_OPENED")
	self:Printf("Will %s refill from the bank.", v and "now" or "no longer")
end

addon.onSets["guildbank"]="OnSet_guildbank"
function addon:OnSet_guildbank(v)
	self[v and "RegisterEvent" or "UnregisterEvent"](self,"GUILDBANKFRAME_OPENED")
	self:Printf("Will %s refill from the guildbank.", v and "now" or "no longer")
end

addon.onSets["overstock"]="OnSet_overstock"
function addon:OnSet_overstock(v)
	self:Printf("Will %s overstock.", v and "now" or "no longer")
end


-- Okay, this is the register-for-an-event-inside-being-LoDed-from-the-same-event bullshit again
-- See http://code.google.com/p/tullamods/issues/detail?id=325
-- and http://www.wowace.com/addons/free-refills/tickets/4-conflict-with-combuctor/
local frame=CreateFrame("Frame", "FreeRefillsInit")
frame:Hide()
frame:SetScript("OnUpdate", function()
	frame:Hide()
	if addon.db.profile.merchant then
		addon:RegisterEvent("MERCHANT_SHOW")
	end
	if addon.db.profile.bank then
		addon:RegisterEvent("BANKFRAME_OPENED")
	end
	if addon.db.profile.guildbank then
		addon:RegisterEvent("GUILDBANKFRAME_OPENED")
	end

	-- In case we get loaded by AddonLoader... (in which case we don't get to see the events ourself)
	if addon.db.profile.bank and select(2,GetContainerNumFreeSlots(BANK_CONTAINER)) then
		addon:BANKFRAME_OPENED()
	end
	-- GLOBALS: GetGuildBankTabPermissions
	if addon.db.profile.guildbank and select(3,GetGuildBankTabPermissions(1)) then
		addon:GUILDBANKFRAME_OPENED()
	end
	if addon.db.profile.merchant and GetMerchantItemInfo(1) then
		addon:MERCHANT_SHOW()
	end
end)

function addon:OnEnable()
	frame:Show()
end

function addon:debug(msg)
	--DBG self:Print("|cff0099CC"..msg.."|r")
end

--[[--------------------------------------------------------------------------------
  Merchant Processing
-----------------------------------------------------------------------------------]]

local lastMerchant=0
function addon:MERCHANT_SHOW()
	self:debug("Merchant Opened")
	if GetTime()<lastMerchant+2 then	-- Prevent taking everything twice when user doubleclicks NPC (items won't have landed in our bags from the first click yet)
		self:debug("Doubleclick? Exiting.")
		return
	end
	lastMerchant=GetTime()
	
	
	for itemID, amount in pairs(self.db.profile.refills) do
		local merchSlot = self:ItemInStock(itemID)
		if merchSlot then
			self:debug("Item Found in Merchant Slot #"..merchSlot)
			local amountInBags = GetItemCount(itemID)
			
			if amountInBags >= amount then
				self:debug("Already have enough in bags.")
			else
				self:debug("Refilling from "..amountInBags.. " to "..amount)
				local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture
					= GetItemInfo(itemID)

				if not itemStackCount then
					-- this happens all the time as of 4.0: self:Print("Err, something odd happened: the item '"..tostring(itemID).."' did not have a stack size?. This could be caused by you having just patched and the WoW item cache not yet having been updated (and in that case is perfectly normal). Try showing the item's tooltip to force an update from the server.")
					self:debug("  Couldn't get stack count for itemID "..tostring(itemID))
					return
				end

				local name, texture, price, quantity, numAvailable, isUsable, extendedCost 
					= GetMerchantItemInfo(merchSlot)
				self:debug("    name:"..tostring(name).." quantity:"..tostring(quantity).." extendedCost:"..tostring(extendedCost))
				if numAvailable<0 then numAvailable=9e9 end
				
				local toBuy = amount-amountInBags
				
				toBuy=min(toBuy,numAvailable)

				while toBuy>0 do
					local now = min(toBuy,itemStackCount)
					-- Non-money purchases can be quirky. not all, but some.
					-- We only buy those according to vendor quantity.
					if extendedCost and now~=quantity then  
						if now<quantity and not addon.db.profile.overstock then
							break
						end
						now=quantity
					end
					self:Printf("  %2u x %s", now, itemLink)
					BuyMerchantItem(merchSlot, now)
					toBuy=toBuy-now
				end
			end
		end
	end
end

function addon:ItemInStock(itemLink)
	self:debug("Searching for Item: "..itemLink)
	local itemName = GetItemInfo(itemLink)	-- becomes nil for uncached items but that's fine, the below check just silently fails
	local numItems = GetMerchantNumItems()
	for i = 1, numItems do
		if GetMerchantItemInfo(i) == itemName then		-- compares names
			return i
		end
	end
end



--[[--------------------------------------------------------------------------------
  Bank Processing
-----------------------------------------------------------------------------------]]

local lastBank = 0
function addon:BANKFRAME_OPENED()
	self:debug("Bank Opened")
	if GetTime()<lastBank+2 then		-- Prevent taking everything twice when user doubleclicks NPC (items won't have landed in our bags from the first click yet)
		self:debug("Doubleclick? Exiting.")
		return
	end
	lastBank=GetTime()
	
	for itemID,amount in pairs(self.db.profile.refills) do
		self:debug(itemID..": "..amount)
		local amountInBags = GetItemCount(itemID)
		local amountInBank = GetItemCount(itemID, true) - amountInBags
		if amountInBank > 0 then
			local amountNeeded = amount - amountInBags
			if amountNeeded<=0 then
				self:debug("Already have enough in bags.")
			else
				if amountNeeded > amountInBank then
					amountNeeded = amountInBank 
				end
				self:debug(amountNeeded.." item(s) to be looted.")
				self:LootBank(itemID, min(amountNeeded, amountInBank))
			end
		end
	end
end

local function linkDecode(link)
	if link then
		return select(3, link:find("Hitem[^|]+|h%[([^[]+)%]"))
	end
end

function addon:LootBank(itemID, amountNeeded)
	self:debug("Bank Looting Logic: "..itemID..": "..amountNeeded)
	
	while amountNeeded>0 do
		local bag,slot,size = LBU:FindSmallestStack("BANK", tonumber(itemID) or tonumber(strmatch(itemID,"item:([0-9]+)")))  -- won't find locked slots
		if not bag then
			self:debug("  - no (more) in bank.")
			break
		end
		self:debug("  found " .. size .. " in "..bag..","..slot)
		
		if size<=amountNeeded then
			self:debug("      looting entire stack")
			-- Don't use this or LBU doesn't know the slot is taken: UseContainerItem(bag,slot)
			PickupContainerItem(bag,slot)
			--DBG local dbag,dslot = 
				LBU:PutItem("BAGS", size)
			--DBG self:debug("  Put it in "..tostring(dbag)..", "..tostring(dslot))
			amountNeeded = amountNeeded - size
		else
			self:debug("      splitting off "..amountNeeded)
			SplitContainerItem(bag, slot, amountNeeded)
			--DBG local dbag,dslot = 
				LBU:PutItem("BAGS", amountNeeded)
			--DBG self:debug("  Put it in "..tostring(dbag)..", "..tostring(dslot))
			amountNeeded = 0
		end
		
	end

end

--[[--------------------------------------------------------------------------------
  Guildbank Processing
-----------------------------------------------------------------------------------]]
-- Ye ye. I know it's a big one and that I need to break it up (esp. around the
-- locals, but lua files are new to me. - Livanamako

-- GLOBALS: GetNumGuildBankTabs GetGuildBankItemLink GetGuildBankItemInfo SplitGuildBankItem PickupGuildBankItem
local lastGuildbank = 0
function addon:GUILDBANKFRAME_OPENED()
	self:debug("Guildbank Opened")
	if GetTime()<lastGuildbank+2 then		-- Prevent taking everything twice when user doubleclicks NPC (items won't have landed in our bags from the first click yet)
		self:debug("Doubleclick? Exiting.")
		return
	end
	lastGuildbank=GetTime()

	for itemID,amount in pairs(self.db.profile.refills) do
		if itemID and amount>0 then
			local amountNeeded = amount-GetItemCount(itemID)

			if amountNeeded>0 then 
				self:debug("Item needed: "..itemID)
				self:debug("Amount needed: "..amountNeeded)
				for tab = 1, GetNumGuildBankTabs() do
					if amountNeeded > 0 then
						local amountTab = 0
						local sortTbl={}
						local index=1
						for slot = 1, 98 do
							local itemLinkGB = GetGuildBankItemLink(tab, slot)
							if itemLinkGB then
								local linkGB = select(2, GetItemInfo(itemLinkGB))
								local linkDB = select(2, GetItemInfo(itemID))
								if linkDB and linkGB and linkDB == linkGB then
									local texture, count, locked = GetGuildBankItemInfo(tab, slot)
									local maxStack = select(8, GetItemInfo(itemID))
									self:debug("Tab: "..tab.." Slot: "..slot.." has "..count.."/"..maxStack)
									amountTab = amountTab + count
									sortTbl[index] = {slot = slot, count = count, maxStack = maxStack}
									index = index + 1
								end
							end
						end
						-- It is guildbank, we want to take big stacks so we find small stacks and get as many needed (remaining or maxStack
						sort(sortTbl, function (a, b) return a.count<b.count end)
						local indexStack = 0
						for indexTbl = 1, #(sortTbl) do
							if amountNeeded>0 then
								local tblItem = sortTbl[indexTbl]
								local slot     = tblItem.slot
								local count    = tblItem.count
								local maxStack = tblItem.maxStack

								if indexStack>0 and sortTbl[indexStack].count>0 then
									local stack = sortTbl[indexStack]
									local scount = stack.count
									local amountCollect = scount + count
									local amountToStack = 0

									if amountCollect<maxStack then
										amountToStack = -scount
									else
										if amountNeeded >= maxStack then 
											amountToStack = maxStack-scount
										elseif amountNeeded > scount then
											amountToStack = amountNeeded - scount
										-- else amountToStack = 0
										end
									end
							
									if amountToStack > 0 then
										self:debug("Tab: "..tab.." moving "..amountToStack.." from "..slot.." to "..stack.slot)
										if amountToStack == count then
											PickupGuildBankItem(tab, slot)
										elseif amountToStack < count then
											SplitGuildBankItem(tab, slot, amountToStack)
										end
										PickupGuildBankItem(tab, stack.slot) -- into stack
									elseif amountToStack<0 then
										self:debug("Tab: "..tab.." moving "..(-amountToStack).." from "..stack.slot.." to "..slot)
										if -amountToStack == scount then
											PickupGuildBankItem(tab, stack.slot)
										elseif -amountToStack < scount then
											SplitGuildBankItem(tab, stack.slot, -amountToStack)
										end
										PickupGuildBankItem(tab, slot) -- into current slot
									end
									sortTbl[indexTbl].count = count - amountToStack
									sortTbl[indexStack].count = scount + amountToStack
								
									count = sortTbl[indexTbl].count
									scount = sortTbl[indexStack].count

									local amountToBags = scount
									if amountNeeded<amountToBags then
										amountToBags = amountNeeded
									end
									if amountToBags>0 then
										if amountToBags == scount then
											PickupGuildBankItem(tab, stack.slot)
										elseif amountToBags < count then
											SplitGuildBankItem(tab, stack.slot, amountToBags)
										end
										self:debug("Tab: "..tab.." moving "..amountToBags.." from "..stack.slot.." to bags")
										LBU:PutItem("BAGS", scount)
										amountNeeded = amountNeeded - scount
									end
								end -- indexStack>0 stack.count>0

								indexStack = indexTbl

								if indexStack == #(sortTbl) then
									local amountToBags = count
									if amountNeeded<amountToBags then
										amountToBags = amountNeeded
									end
									if amountToBags>0 then
										if amountToBags == count then
											PickupGuildBankItem(tab, slot)
										elseif amountToBags < count then
											SplitGuildBankItem(tab, slot, amountToBags)
										end
										self:debug("Tab: "..tab.." moving "..amountToBags.." from "..slot.." to bags")
										LBU:PutItem("BAGS", amountToBags)
										amountNeeded = amountNeeded - amountToBags
									end
								end

							end -- amountNeeded > 0
						end -- for indexTbl		
					end -- amountNeeded > 0
				end -- for tab
			end -- amountNeeded > 0
		end -- if itemID and amount > 0
	end -- for itemID, amount in "db...refill"
end

--[[--------------------------------------------------------------------------------
  Command Handlers
-----------------------------------------------------------------------------------]]

local function setrefill(link,num)
	-- itemstrings now have the player's CURRENT level in them, so we have to take the itemstring apart for matching
	local substr = strmatch(link, "item:%d+:") or "["..link.."]"
	for k,v in pairs(addon.db.profile.refills) do
		if strfind(k, substr, 1, true) then
			addon.db.profile.refills[k] = num
			return k
		end
	end
	if num then
		addon.db.profile.refills[link] = num
	end
end

function addon:AddRefillItem(info, msg)
	--add an item to the refill list
	local link, num = strmatch(msg,"^%s*(|c.+|Hitem.+)%s(%d+)%s*$")
	if not link or not num then
		self:Print(L["Invalid input; input must be '<item> #'."])
		return
	end
	num = tonumber(num)
	setrefill(link,num)
	self:Printf(L["Added %s - Amount:%d"], link, num)
end

local function GetStackSize(item)
	local itemStack = select(8, GetItemInfo(item))
	return itemStack
end

function addon:AddRefillItemStack(info, msg)
	--add an item to the refill list by stack
	local link, num = strmatch(msg,"^%s*(|c.+|Hitem.+)%s(%d+)%s*$")
	if not link or not num then
		self:Print(L["Invalid input; input must be '<item> #'."])
		return
	end
	num = tonumber(num)
	num = num * GetStackSize(link)
	setrefill(link,num)
	self:Printf(L["Added %s - Amount:%d"], link, num)
end

function addon:RemoveRefillItem(info, msg)
	--remove an item from the refill list
	local item = setrefill(msg, nil)
	if item then
		self:Printf(L["Removed %s."], item)
	else
		self:Printf(L["Could not find '%s'."], msg)
	end
end

function addon:ClearRefills()
	--clear the refill list
	self.db.profile.refills = {}
	self:Print(L["Item List Cleared"])
end

local function sortLink(a,b)
	return strmatch(a,"|h%[(.*)%]|h") < strmatch(b,"|h%[(.*)%]|h")
end

function addon:List()
	--report back which items are currently on the refill list, and their quanitity
	self:Print(L["Refill List: "])
	local sorted = {}
	for itemID in pairs(self.db.profile.refills) do
		tinsert(sorted, itemID)
	end
	sort(sorted,sortLink)
	
	for _,itemID in ipairs(sorted) do
		self:Print("     "..itemID.." ("..self.db.profile.refills[itemID]..")")
	end
	if not next(self.db.profile.refills) then 
		self:Print(L["     None"]) 
	end
end


